// ZenLib::Server::Http::Dispatcher - HTTP server
// Copyright (C) 2008-2008 Jerome Martinez, Zen@MediaArea.net
//
// This software is provided 'as-is', without any express or implied
// warranty.  In no event will the authors be held liable for any damages
// arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it
// freely, subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented; you must not
//    claim that you wrote the original software. If you use this software
//    In a product, an acknowledgment In the product documentation would be
//    appreciated but is not required.
// 2. Altered source versions must be plainly marked as such, and must not be
//    misrepresented as being the original software.
// 3. This notice may not be removed or altered from any source distribution.
//
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

//---------------------------------------------------------------------------
#include "ZenLib/Conf_Internal.h"
#ifdef __BORLANDC__
    #pragma hdrstop
#endif
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
#include "ZenLib/Server/Http/Dispatcher.h"
#include "ZenLib/Format/Http/Http_Utils.h"
#include "ZenLib/Ztring.h"
#include <sstream>
using namespace std;
using namespace ZenLib::Format::Http;
//---------------------------------------------------------------------------

namespace ZenLib
{

namespace Server
{

namespace Http
{

//***************************************************************************
// Constructor/Destructor
//***************************************************************************

//---------------------------------------------------------------------------
Dispatcher::Dispatcher()
{
    Reqs=NULL;
}

//---------------------------------------------------------------------------
Dispatcher::Dispatcher(vector<ZenLib::Format::Http::Request*>* Main_Requests)
{
    Reqs=Main_Requests;
    Reqs_Pos=0;
    set_max_connections(Main_Requests->size());
}

//***************************************************************************
// Hooks from dlib
//***************************************************************************

//---------------------------------------------------------------------------
void Dispatcher::on_connect (dlib::connection& con)
{
    dlib::server::ssbuf1a buf(&con);
    std::istream In(&buf);
    std::ostream Out(&buf);

    Request* Req=NULL;

    try
    {
        //Searching an available slot
        Req=Reqs->at(Reqs_Pos);
        if (Req->IsUsed)
        {
            //No handlers are available!
            Out << "HTTP/1.0 503\r\n";
            Out << "\r\n";
            Out << "HTTP: Too many connections";
            return;
        }
        Req->IsUsed=true;
        
        //Positionning for the next time
        Reqs_Pos++;
        if (Reqs_Pos>=Reqs->size())
            Reqs_Pos=0;

        //First line, "Method Path Norm"
        //-Method
        string Method;
        In>>Method; 
             if (Method.size()==3 && Method[0]=='G' && Method[1]=='E' && Method[2]=='T')
            ;
        else if (Method.size()==4 && Method[0]=='P' && Method[1]=='O' && Method[2]=='S' && Method[3]=='T')
            ;
        else if (Method.size()==4 && Method[0]=='H' && Method[1]=='E' && Method[2]=='A' && Method[2]=='D')
            Req->Http->HeadersOnly=true;
        else
        {
            Out << "HTTP/1.0 501\r\n";
            Out << "\r\n";
            return; //Unknown request
        }
        //-Path
        In>>Req->Http->Path;
        if (Req->Http->Path.empty() || Req->Http->Path[0]!='/')
        {
            Out << "HTTP/1.0 404\r\n";
            Out << "\r\n";
            return; //Problem
        }
        if (Req->Http->Path.find("%")!=string::npos)
            Req->Http->Path=Format::Http::URL_Encoded_Decode(Req->Http->Path);
        //-Norm
        string Line;
        getline(In, Line); //Drop the first line, no more needed
        
        //Headers
        do
        {
            //Getting the line
            getline(In, Line);
            if (!Line.empty() && Line[Line.size()-1]=='\r')
                Line.resize(Line.size()-1); //Remove the \r
            
            //Processing the line, form is "aaa: bbb"
            if (!Line.empty())
            {
                string::size_type Separator_Pos=Line.find(':');
                string::size_type Content_Begin=Line.find_first_not_of(' ', Separator_Pos+1);
                if (Separator_Pos!=string::npos && Content_Begin!=string::npos)
                {
                    Ztring Command=Line.substr(0, Separator_Pos);
                    if (Command=="Cookie")
                    {
                        Ztring Cookie=Line.substr(Separator_Pos+1, string::npos);
                        while (!Cookie.empty())
                        {
                            string::size_type Cookie_Pos=Cookie.rfind(';');
                            if (Cookie_Pos==string::npos)
                                Cookie_Pos=(string::size_type)-1;
                            Ztring Line=Cookie.substr(Cookie_Pos+1, string::npos);
                            Line.TrimLeft(' ');
                            if (Cookie_Pos!=(string::size_type)-1)
                            {
                                Cookie.resize(Cookie_Pos);
                                Cookie.TrimLeft(' ');
                            }
                            else
                                Cookie.clear();
                                
                            string::size_type Separator_Pos=Line.find('=');
                            if (Separator_Pos!=string::npos)
                                Req->Http->Request_Cookies[Line.substr(0, Separator_Pos)]=Format::Http::URL_Encoded_Decode(Line.substr(Separator_Pos+1, string::npos));
                        }
                    }
                    else
                        Req->Http->Request_Headers[Command]=Line.substr(Content_Begin, string::npos);
                }
            }
        }
        while (!Line.empty());

        //Info
        if (Method.size()==3 && Method[0]=='G' && Method[1]=='E' && Method[2]=='T'
         || Method.size()==4 && Method[0]=='P' && Method[1]=='O' && Method[2]=='S' && Method[3]=='T')
        {
            do
            {
                Ztring Content; 
                
                //Getting the line
                string::size_type Interogation_Pos=Req->Http->Path.find('?');
                if (Interogation_Pos!=string::npos)
                {
                    Content=Req->Http->Path.substr(Interogation_Pos+1, string::npos);
                    Req->Http->Path.resize(Interogation_Pos);
                }
                
                if (Method.size()==4) //Only for POST
                {
                    int64u Content_Lengh=0;
                    std::map<std::string, std::string>::iterator Header_Content_Lengh_Element=Req->Http->Request_Headers.find("Content-Length");
                    if (Header_Content_Lengh_Element==Req->Http->Request_Headers.end())
                        Header_Content_Lengh_Element=Req->Http->Request_Headers.find("Content-length");
                    if (Header_Content_Lengh_Element!=Req->Http->Request_Headers.end())
                        Content_Lengh=Ztring(Header_Content_Lengh_Element->second).To_int64u();
                    if (Content_Lengh>1024) //Verifying there is no big element
                    {
                        Out << "HTTP/1.0 501\r\n";
                        Out << "\r\n";
                        return; //Malformed request
                    }

                    size_t Content_Size_Current=Content.size();
                    if (Content_Size_Current)
                    {
                        Content+='&';
                        Content_Size_Current++;
                    }
                    Content.resize(Content_Size_Current+(size_t)Content_Lengh+1);
                    In.read(&Content[Content_Size_Current], (size_t)Content_Lengh);
                    Content[Content_Size_Current+(size_t)Content_Lengh]='\0';
                }
                
                //Processing the line, form is "aaa=bbb&..."
                while (!Content.empty())
                {
                    string::size_type Content_Pos=Content.rfind('&');
                    if (Content_Pos==string::npos)
                        Content_Pos=(string::size_type)-1;
                    Ztring Line=Content.substr(Content_Pos+1, string::npos);
                    if (Content_Pos!=(string::size_type)-1)
                        Content.resize(Content_Pos);
                    else
                        Content.clear();
                        
                    string::size_type Separator_Pos=Line.find('=');
                    if (Separator_Pos!=string::npos)
                        Req->Http->Request_Queries[Line.substr(0, Separator_Pos)]=Format::Http::URL_Encoded_Decode(Line.substr(Separator_Pos+1, string::npos));
                }
            }
            while (!Line.empty());
        }

        //From socket
        Req->Http->Foreign_IP=con.get_foreign_ip();
        Req->Http->Local_IP=con.get_local_ip();
        Req->Http->Foreign_Port=con.get_foreign_port();
        Req->Http->Local_Port=con.get_local_port();
        
        //Request
        Req->Manage();

        //Preparing headers
        Out << "HTTP/1.0 "<< Req->Http->Response_HTTP_Code << "\r\n";
        for (std::map<std::string, std::string>::iterator Temp=Req->Http->Response_Headers.begin(); Temp!=Req->Http->Response_Headers.end(); Temp++)
            Out << Temp->first << ": " << Temp->second << "\r\n";
        Req->Http->Response_Cookies.Create_Lines(Out);
        std::map<std::string, std::string>::iterator Content_Type_Element=Req->Http->Response_Headers.find("Content-Type");
        if (Content_Type_Element!=Req->Http->Response_Headers.end())
            Out << "Content-Type: "<< Content_Type_Element->second << "\r\n";
        else if (Req->Http->Response_HTTP_Code==200)
        {
            if (!Req->Http->Response_Body.empty() && Req->Http->Response_Body[0]=='<')
                Out << "Content-Type: "<< "text/html; charset=utf-8" << "\r\n";
        }
        if (!Req->Http->Response_Body.empty())
            Out << "Content-Length: " << Req->Http->Response_Body.size() << "\r\n";
        Out << "\r\n";
        if (!Req->Http->HeadersOnly && !Req->Http->Response_Body.empty())
            Out << Req->Http->Response_Body.c_str();
    }
    
    catch (std::exception&)
    {
        Out << "HTTP/1.0 503\r\n";
        Out << "\r\n";
    }

    Req->IsUsed=false;
}

} //Namespace

} //Namespace

} //Namespace
